/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster File Systems, Inc.
 * Copyright (C) 2001 Tacit Networks, Inc. <phil@off.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <glib.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <sys/socket.h>
#include <intersync.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>

#include <linux/intermezzo_fs.h>
#include <linux/intermezzo_idl.h>
#include <izolib.h>

#if 0
static int is_http_to_errno(int status)
{
        int err = 0;

        if (status >= 200 && status <= 300) 
                err = 0;
        else if (status == 404)
                err = ENOENT;
        else if (status == 403)
                err = EPERM;
        else if (status < 0)
                err = status;
        else
                err = EIO;      /* What should be the catchall? */

        return err;
}
#endif

static int is_http_get(struct is_context *ctxt, char *uri, int start_byte,
                int end_byte)
{
        int rc;

        if (ctxt->ctxt_req == NULL) {
                CERROR("No connection, what happened?\n");
                return ENOTCONN;
        }

        CDEBUG(D_HTTP, "Fetching uri '%s' (range %d-%d)\n", uri, start_byte,
                        end_byte);

        rc = httpclient_fetch(ctxt->ctxt_req, uri, start_byte, end_byte);
        if (rc) {
                CDEBUG(D_HTTP, "httpclient_fetch() failed: %s\n",
                                httpclient_strerror(rc));
        }

        return httpclient_errno(rc);
}

int is_get_file(struct is_context *ctxt, char *path,
                struct lento_vfs_context *info)
{
        char *uri, *file;
        FILE *fp;
        int rc;
        struct is_peer *peer = is_ctxt_peer(ctxt); 
        struct is_fileset *fset = ctxt->ctxt_replicator->rp_fset;

        if (!info || info->remote_ino == 0)
                uri = g_strdup_printf("http://%s:%d/cache/.intermezzo/%s/ROOT"
                                      "/%s", peer->srv_name, peer->srv_port,
                                      fset->fset_name, path);
        else
                uri = g_strdup_printf("http://%s:%d/cache/..iopen../...ino:%Lx:%Lx",
                                      peer->srv_name, peer->srv_port, 
                                      info->remote_ino, 
                                      info->remote_generation);

        file = g_strdup_printf("%s%s", fset->fset_mtpt, path);
        fp = fopen(file, "w");
        if (fp == NULL) {
                CERROR("backfetch failed; local open(%s): %s\n", file,
                                g_strerror(errno));
                rc = -1;
        }
        else {
                rc = httpclient_fetch_file(ctxt->ctxt_req, uri, 0, 0, fp);
                if (rc) {
                        CDEBUG(D_HTTP, "httpclient_fetch() failed: %s\n",
                                        httpclient_strerror(rc));
                        rc = httpclient_errno(rc);
                }
        }
        g_free(file);
        g_free(uri);
        
        if (rc)
                return rc;

        /* at this point, we've successfully xferred the file. */

        if (info && info->flags) {
                CDEBUG(D_INFO, "adjust lml, flags %x\n", info->flags);
                rc = izo_ioc_adjust_lml(fileno(fp), info);
                if (rc < 0)
                        is_die("kernel cannot adjust lml: %s (%d)\n",
                                        g_strerror(errno), errno); 
        }
        fclose(fp);

        return rc;
}


/*
 * Make all the directories in the path, 
 * and create an empty file.
 */
static int branch_fixpath(struct is_fileset *fset, struct kml_rec *rec) 
{
        char *dir, *path, *file;
        int fd;

        dir = izo_dirname(rec->path);
        if (dir == NULL)
                return -1;
        if (strlen(dir) > 1) {
                int rc;
                path = g_strdup_printf("%s%s", fset->fset_mtpt, dir);
                if (path == NULL) {
                        g_free(dir);
                        return -1;
                }
                CDEBUG(D_INFO, "creating directory %s\n", path);
                rc = izo_mkpath(path, 0777);
                if (rc < 0) {
                        CERROR("unable to build dir %s\n", path);
                }
                g_free(path);
        }
        g_free(dir);

        file = g_strdup_printf("%s%s", fset->fset_mtpt, rec->path);
        if (file == NULL) {
                return -1;
        }
        CDEBUG(D_INFO, "createing file: %s\n", file);
        fd = open(file, O_CREAT|O_RDONLY, 0666);
        if (fd < 0) {
                CERROR("Unable to open %s: %s\n", file, g_strerror(errno));
        } else 
                close(fd);

        g_free(file);
        return 0;
        
}

/*#define DEBUG_KML_OFFSET*/
static int is_get_kml_range(struct is_context *ctxt, char *branch,
                            int start_byte, int end_byte)
{
        char *uri;
        struct kml_rec rec;
        char *buf, *ptr, *end;
        struct is_peer *peer = ctxt->ctxt_replicator->rp_peer;
        struct is_fileset *fset = ctxt->ctxt_replicator->rp_fset;
        int rc;
        __u32 server = 0;
        __u32 isbranch = 0;
        time_t start_time, end_time;
        
        char *tmp_buf = NULL;          /* unpack using a copy, so 
                                   kernel can upack again.  */
        char *tmp_ptr, *tmp_end;

#ifdef DEBUG_KML_OFFSET
        struct izo_rcvd_rec lr_rec;
        __u64 kml_remote_offset = 0;
        long kml_rec_size = 0;  
#endif
        if (start_byte && end_byte && start_byte >= end_byte)
                return 0;

        if (fset->fset_server == NULL)
                /* We are the server */
                server = 1;
        if (fset->fset_parent != NULL)
                isbranch = 1;

        uri = g_strdup_printf("http://%s:%d/cache/.intermezzo/%s/kml",
                              peer->srv_name, peer->srv_port,
                              fset->fset_name);
        rc = is_http_get(ctxt, uri, start_byte, end_byte);
        g_free(uri);

        if (rc)
                return rc;
        if (ctxt->ctxt_req == NULL ||
            ctxt->ctxt_req->len == 0) {
                CERROR("No KML?\n");
                return ENOENT;
        }

        ptr = buf = ctxt->ctxt_req->body;
        end = buf + ctxt->ctxt_req->len;

        CDEBUG(D_KML, "kml size: %d\n", end - ptr);

        tmp_ptr = tmp_buf = g_malloc(end - ptr);
        memcpy(tmp_buf, buf, end-ptr);
        tmp_end = tmp_buf + (end-ptr);

#ifdef DEBUG_KML_OFFSET
        rc = izo_ioc_get_rcvd(fset->fset_mtpt_fd, peer->srv_uuid, &lr_rec);
        if (rc == -EINVAL) {
                /* We have no last_rcvd data for this peer, do not panic. */
                memset(&lr_rec, 0, sizeof(lr_rec));
                rc = 0;
        }
        if (rc < 0) {
                CDEBUG(D_INFO, "get_rcvd failure: %d, %d\n", rc, errno);
                goto out_free;
        }
        if (start_byte != lr_rec.lr_remote_offset)
                CERROR("start_byte %d != remote_offset %Lu\n", 
                       start_byte, 
                       lr_rec.lr_remote_offset);
        kml_remote_offset = start_byte;
#endif
        CDEBUG(D_REINT, "reint start: %d\n", end - buf);
        start_time = time(NULL);
        while (ptr < end) {
                int rc;
                char *old_ptr = tmp_ptr;
                
                if ((rc = kml_unpack(&rec, &tmp_ptr, tmp_end)) != 0) {
                        CERROR("kml_unpack failed: %s\n", g_strerror(-rc));
                        rc = -1;
                        goto out_free;
                } else {
                        char *str = kml_print_rec(&rec, 1);
                        CDEBUG(D_KML, "%s\n", str);
                        free(str);
                        if (isbranch && server 
                            && rec.prefix.hdr->opcode == KML_OPCODE_CLOSE) {
                                branch_fixpath(fset, &rec);
                        }
                        rc = kml_iocreint(tmp_ptr - old_ptr, ptr,
                                          start_byte,
                                          fset->fset_mtpt_fd,
                                          peer->srv_uuid, server);
                        if (rc == ENOTCONN) {
                                CERROR("disconnected, reintegration halted\n");
                                goto out_free;
                        } else if (rc) {
                                is_die("reintegrate failed, aborting: %s\n",
                                          g_strerror(rc));
                        }
                        /* Success?  Set the last_rcvd fields. */
                        /* FIXME: Set them _all_. */
                        ctxt->ctxt_replicator->rp_rcvd_rec.lr_remote_offset +=
                                tmp_ptr - old_ptr;
                        start_byte += tmp_ptr - old_ptr;

                        ptr += tmp_ptr - old_ptr;
#ifdef DEBUG_KML_OFFSET
                        kml_rec_size += rec.suffix->len;
                        /* After each reintr, check offset and make sure it is correct */
                        rc = izo_ioc_get_rcvd(fset->fset_mtpt_fd, peer->srv_uuid, &lr_rec);
                        if (rc == -EINVAL) {
                                /* We have no last_rcvd data for this peer, do not panic. */
                                memset(&lr_rec, 0, sizeof(lr_rec));
                                rc = 0;
                        }
                        if (rc < 0) {
                                CERROR("get_rcvd failure: %d, %d\n", rc, errno);
                                goto out_free;
                                
                        }
                        kml_remote_offset += tmp_ptr - old_ptr;
                        if (kml_remote_offset != lr_rec.lr_remote_offset) {
                                fflush(stdout);
                                fflush(stderr);
                                is_die("offsets don't match! %Lu != %Lu\n",
                                        kml_remote_offset, lr_rec.lr_remote_offset);
                        }
#endif
                        
                }
        }
#ifdef DEBUG_KML_OFFSET
        CERROR("Actual KML bytes read %ld\n", kml_rec_size);
        CERROR("Final offset should be %Lu\n:", kml_remote_offset);
#endif
        end_time = time(NULL);
        CDEBUG(D_REINT, "reint complete: %d in %ld secs\n", end - buf, 
               end_time - start_time);
        rc = 0;
 out_free:
        g_free(tmp_buf);
        return rc;
}

static int is_remote_cgi(struct is_context *ctxt, char *name)
{
        char *uri;
        int rc;
        struct is_peer *peer;

        peer = is_ctxt_peer(ctxt); 

        uri = g_strdup_printf("http://%s:%d/cgi-bin/cgi_%s",
                              peer->srv_name, peer->srv_port, name);
        rc = is_http_get(ctxt, uri, 0, 0);
        g_free(uri);

        return rc;
}

static int is_get_repstatus(struct is_context *ctxt, __u64 local_kmlsize,
                            __u64 *remote_kmlsize,
                            struct izo_rcvd_rec *rec)
{
        char uuidstr[UUID_UNPARSED_SIZE], *name, *body, *end;
        struct is_fileset *fset = ctxt->ctxt_replicator->rp_fset;
        struct izo_rcvd_rec *lr_client = &ctxt->ctxt_replicator->rp_rcvd_rec;
        int rc, len;

        uuid_unparse(ctxt->ctxt_intersync->is_local_uuid, uuidstr);
        name = g_strdup_printf("repstatus?fset=%s&uuid=%s&kml_size=%Lu"
                               "&kml_start=%Lu&kml_lastseen=%Lu"
                               "&kml_start_recno=%Lu&kml_lastseen_recno=%Lu",
                               fset->fset_name, uuidstr, local_kmlsize,
                               lr_client->lr_local_offset, lr_client->lr_remote_offset,
                               lr_client->lr_local_recno, lr_client->lr_remote_recno);
        rc = is_remote_cgi(ctxt, name);
        g_free(name);
        if (rc)
                return rc;

        body = ctxt->ctxt_req->body;
        len = ctxt->ctxt_req->len;

        if (len == 0)
                return -1;

        end = body + len;
        LUNLOGV(*remote_kmlsize, __u64, body, end);

        /* FIXME: none of the macros do this type of unpacking! */
        if (body + sizeof(rec->lr_uuid) > end)
                return -EFAULT;
        memcpy(rec->lr_uuid, body, sizeof(rec->lr_uuid));
        body += sizeof(rec->lr_uuid);

        LUNLOGV(rec->lr_remote_recno, __u64, body, end);
        LUNLOGV(rec->lr_remote_offset, __u64, body, end);
        LUNLOGV(rec->lr_local_recno, __u64, body, end);
        LUNLOGV(rec->lr_local_offset, __u64, body, end);
        LUNLOGV(rec->lr_last_ctime, __u64, body, end);

        return 0;
}

int is_wait_for_kml(struct is_fileset *fset)
{
        g_mutex_lock(fset->fset_kml_mutex);
        while (fset->fset_is_fetching_kml) {
                CDEBUG(D_KML,"waiting for kml to finish\n");
                g_cond_wait(fset->fset_kml_cond, fset->fset_kml_mutex);
        }
        g_mutex_unlock(fset->fset_kml_mutex);
        CDEBUG(D_KML, "kml is complete\n");
        return 0;
}

int is_get_kml(struct is_context *ctxt, __u64 kmlsize)
{
        int rc;
        struct is_peer *peer;
        struct is_fileset *fset = ctxt->ctxt_replicator->rp_fset;
        struct izo_rcvd_rec *rec = &ctxt->ctxt_replicator->rp_rcvd_rec;
        static int already = 0;        /* count calls when already fetching  */

        g_mutex_lock(fset->fset_kml_mutex);
        if (fset->fset_is_fetching_kml) {
                CDEBUG(D_CHAT, "We're already fetching KML for this fileset in another "
                       "thread; skipping.\n");
                already++;
                if (kmlsize && kmlsize < fset->fset_remote_kml_size)
                        CDEBUG(D_KML, "BUG?  kmlsize < fset_remote_kml_size!! "
                               "(%Lu < %Lu)\n", kmlsize,
                               fset->fset_remote_kml_size);
                if (kmlsize)
                        fset->fset_remote_kml_size = kmlsize;
                g_mutex_unlock(fset->fset_kml_mutex);
                return 0;
        }
        already = 0;
        fset->fset_is_fetching_kml = 1;
 restart:
        g_mutex_unlock(fset->fset_kml_mutex);

        peer = is_ctxt_peer(ctxt);

        CDEBUG(D_KML, "remote kmlsize: %Lu, local kmlsize: %Lu, already: %d\n",
               kmlsize, rec->lr_remote_offset, already);

        if (kmlsize < rec->lr_remote_offset) {
                CDEBUG(D_KML, "Remote kmlsize is smaller than lr_remote_offset! "
                          "Ignoring.\n");
                rc = 0;
                goto out;
        }

        rc = is_get_kml_range(ctxt, NULL, rec->lr_remote_offset, kmlsize-1);

 out:
        g_mutex_lock(fset->fset_kml_mutex);
        if (rc == 0 && fset->fset_remote_kml_size > kmlsize) {
                /* We've been asked to fetch more KML */
                kmlsize = fset->fset_remote_kml_size;
                goto restart;
        }
        fset->fset_is_fetching_kml = 0;
        g_cond_broadcast(fset->fset_kml_cond);
        g_mutex_unlock(fset->fset_kml_mutex);
        return rc;
}

/* Check to see that only one side (or neither side) updated the KML since we
 * last checked in.  If both updated the KML, we have a possible conflict
 * situation which we currently don't handle. */
int is_check_offsets(struct is_context *ctxt)
{
        int rc;
        struct intersync *is = ctxt->ctxt_intersync;
        struct is_replicator *rep = ctxt->ctxt_replicator;
        struct is_fileset *fset = rep->rp_fset;
        struct izo_rcvd_rec remote_rec;
        struct izo_rcvd_rec *lr_client = &rep->rp_rcvd_rec;
        struct izo_rcvd_rec *lr_server = &remote_rec;
        __u64 client_kmlsize, server_kmlsize;

        rc = izo_ioc_get_kml_size(fset->fset_mtpt_fd, &client_kmlsize);
        if (rc < 0) {
                CERROR("no local kml size\n");
                return rc;
        }

        rc = is_get_repstatus(ctxt, client_kmlsize, &server_kmlsize, lr_server);
        if (rc) {
                CERROR("no remote repstatus\n");
                return rc;
        }

        CDEBUG(D_KML, "check_offsets: \n"
               "our kml size %Lu, server last fetched %Lu.\n"
               "Server kml size %Lu, we last fetched %Lu;\n"
               "Server suggests we fetch from  %Lu.\n",
               client_kmlsize, lr_server->lr_remote_offset,
               server_kmlsize, lr_client->lr_remote_offset,
               lr_server->lr_local_offset);

        /* If the server knows about more kml than we do, use it */
        if (lr_server->lr_local_offset > lr_client->lr_remote_offset)
                lr_client->lr_remote_offset = lr_server->lr_local_offset;

        if (client_kmlsize > lr_server->lr_remote_offset)
                CDEBUG(D_KML, "Notice: our KML has grown since we last saw peer\n");
        if (server_kmlsize > lr_client->lr_remote_offset)
                CDEBUG(D_KML, "Notice: the KML of peer has grown since we last saw "
                       "it\n");

        /* If both kml files have grown, there may be conflicts */
        if (client_kmlsize > lr_server->lr_remote_offset &&
            server_kmlsize > lr_client->lr_remote_offset)
                g_error("peer possibly in conflict, cannot continue.\n");

        /* If server has grown, fetch */
        if (server_kmlsize > lr_client->lr_remote_offset) {
                __u64 *kmlsize = g_malloc(sizeof(__u64));
                *kmlsize = server_kmlsize;
                rc =  is_run_job(is, CONTEXT_JOB_KML, NULL, rep, kmlsize);
                if (rc == ENOTCONN) {
                        CERROR("KML reint interrupted\n");
                } else if (rc) {
                        is_die("KML reint error: %s\n", g_strerror(rc));
                }
        }

        /* If client kml has grown, tell server to fetch */
        if (client_kmlsize > lr_server->lr_remote_offset) {
                /* Must have the permit before GO_FETCH_KML. */
                if(is_run_job(is, CONTEXT_JOB_GET_PERMIT, NULL, rep, NULL) != 0)
                        g_error("Initial GET PERMIT failed.\n");
                fset->fset_permit_holder = NULL;
                
                if (is_run_job(is, CONTEXT_JOB_GO_FETCH_KML, NULL, rep, NULL) != 0)
                        g_error("Initial GO_FETCH_KML failed.\n");
        }

        return 0;
}

int is_get_uuid(struct is_context *ctxt, struct is_get_uuid_data * uuid_data)
{
        char *body;
        struct is_peer *peer;
        int rc, len;
        char uuidstr[UUID_UNPARSED_SIZE];
        char *cgi;

        CERROR("\n");
        uuid_unparse(ctxt->ctxt_intersync->is_local_uuid, uuidstr);
        cgi = g_strdup_printf("uuid?client_port=%llu&uuid=%s&client_flag=%d", 
                              uuid_data->port, 
                              uuidstr,
                              uuid_data->client_flag);
        rc = is_remote_cgi(ctxt, cgi);
        g_free(cgi);
        if (rc) {
                CERROR("error fetching UUID %d\n", rc);
                return rc;
        }

        body = ctxt->ctxt_req->body;
        len = ctxt->ctxt_req->len;

        if (len == 0) {
                CERROR("empty result when UUID %d\n", rc);
                return -1;
        }

        CDEBUG(D_INFO, "body: %*s\n", len, body);
        peer = is_ctxt_peer(ctxt);

        strncpy(uuidstr, body, sizeof(uuidstr) - 1);
        uuidstr[sizeof(uuidstr) - 1] = '\0';
        if (uuid_parse(uuidstr, peer->srv_uuid) != 0) {
                CERROR("UUID is not valid.\n");
                return -1;
        }

        g_mutex_lock(peer->srv_cxn_mutex);
        peer->srv_connected = 1;
        g_mutex_unlock(peer->srv_cxn_mutex);

        return 0;
}

int is_get_fileid(struct is_context *ctxt, struct is_fileid_data *fid_data)
{
        char *body;
        int len;
        int fd, rc = 0;
        char out[4096];
        struct is_fileset *fset = ctxt->ctxt_replicator->rp_fset;
        __u64 ino, generation;
        char *cgi;

        izo_to_base64(out, sizeof(out), fid_data->rec, fid_data->reclen);
        cgi = g_strdup_printf("fileid?fsetname=%s&rec=%s",
                              fset->fset_name,
                              out);
        rc = is_remote_cgi(ctxt, cgi);
        g_free(cgi);
        if (rc) {
                CERROR("error fetching UUID %d\n", rc);
                return rc;
        }

        body = ctxt->ctxt_req->body;
        len = ctxt->ctxt_req->len;
        CDEBUG(D_INFO, "fileid response: '%s'\n", body);
        if (!body) {
                return EINVAL;
        }
        rc = sscanf(body, "%Lx:%Lx", &ino, &generation);
        if (rc != 2) {
                CERROR("scanf failed, no ino data\n");
                return EINVAL;
        } 
        CDEBUG(D_INFO,"setting fileid %s:%s, %Lx:%Lx\n", fset->fset_mtpt,
               fid_data->path, ino, generation);
        rc = izo_ioc_set_fileid(fset->fset_mtpt_fd, fid_data->path, ino, generation);
        CDEBUG(D_INFO,"setting fileid status %d\n", rc);
        close(fd);

        return rc;
}
int is_get_profile(struct is_context *ctxt)
{
        char *uri, *fset;
        struct is_peer *peer = is_ctxt_peer(ctxt);
        int rc;

        uri = g_strdup_printf("http://%s:%d/profile",
                              peer->srv_name, peer->srv_port);
        rc = is_http_get(ctxt, uri, 0, 0);
        g_free(uri);
        
        if (rc)
                return rc;
        
        if (ctxt->ctxt_req == NULL || ctxt->ctxt_req->len == 0) {
                CERROR("no body in profile!\n");
                return 0;
        }

        fset = ctxt->ctxt_req->body;
        if (fset[strlen(fset) - 1] == '\n')
                fset[strlen(fset) - 1] = '\0';

        peer->srv_cell->cell_root_fset_name = g_strdup(fset);
        return 0;
}

int is_get_permit(struct is_context *ctxt)
{
        char *name, *body;
        int rc, len;
        struct is_fileset *fset = ctxt->ctxt_replicator->rp_fset;
        char uuidstr[UUID_UNPARSED_SIZE];
        char *s;

        uuid_unparse(ctxt->ctxt_intersync->is_local_uuid, uuidstr);
        name = g_strdup_printf("permit?fset=%s&uuid=%s",
                               fset->fset_name, uuidstr);
        rc = is_remote_cgi(ctxt, name);
        g_free(name);
        if (rc)
                return rc;

        len = ctxt->ctxt_req->len;
        if (len == 0)
                return -1;

        body = ctxt->ctxt_req->body;

        /* the other side returns zero to indicate success or
         * an error code. */
        s = g_malloc(len + 1);
        memcpy(s, body, len);
        s[len] = '\0';
        rc = strtol(s, NULL, 0);
        g_free(s);

        if (rc)
                CERROR("Remote returned permit failure: %d\n", rc);

        return rc;
}

/*
 * Do the check if fetch_pending before queuing thread. This should not
 * block for long, and keeps the peer->srv_kml_size  up to date faster.
 */
int is_go_fetch_notify(struct intersync *is, struct is_replicator *rep, __u64 *kmlsize)
{
        struct is_peer *peer = rep->rp_peer;

        g_mutex_lock(peer->srv_kml_mutex);
        if (peer->srv_go_fetch_pending) {
                CDEBUG(D_CHAT, "We're already doing a 'go_fetch_kml' on this peer in another thread; skipping.\n");
                if (*kmlsize < peer->srv_kml_size)
                        CDEBUG(D_KML, "BUG?  kmlsize < srv_kml_size!! (%Lu < %Lu)\n",
                               *kmlsize, peer->srv_kml_size);
                else
                        peer->srv_kml_size = *kmlsize;
                g_mutex_unlock(peer->srv_kml_mutex);
                return 0;
        }
        g_mutex_unlock(peer->srv_kml_mutex);
        is_queue_job(is, CONTEXT_JOB_GO_FETCH_KML, NULL, rep, kmlsize);
        return 0;

}

int is_go_fetch_kml(struct is_context *ctxt, __u64 kmlsize)
{
        char *name;
        int rc;
        struct is_peer *peer = is_ctxt_peer(ctxt);
        struct is_fileset *fset = ctxt->ctxt_replicator->rp_fset;
        static int already;
        char uuidstr[UUID_UNPARSED_SIZE];

        if (kmlsize == 0) {
                rc = izo_ioc_get_kml_size(fset->fset_mtpt_fd, &kmlsize);
                if (rc < 0)
                        return rc;
        }

        /* go fetch notify should take care of this, but just in case */
        g_mutex_lock(peer->srv_kml_mutex);
        if (peer->srv_go_fetch_pending) {
                CDEBUG(D_KML, "We're already doing a 'go_fetch_kml' on this peer in another thread; skipping.\n");
                already++;
                if (kmlsize < peer->srv_kml_size)
                        CDEBUG(D_KML, "BUG?  kmlsize < srv_kml_size!! (%Lu < %Lu)\n",
                               kmlsize, peer->srv_kml_size);
                else
                        peer->srv_kml_size = kmlsize;
                g_mutex_unlock(peer->srv_kml_mutex);
                return 0;
        }
        already = 0;
        peer->srv_go_fetch_pending = 1;
 restart:
        g_mutex_unlock(peer->srv_kml_mutex);

        uuid_unparse(ctxt->ctxt_intersync->is_local_uuid, uuidstr);
        name = g_strdup_printf("go_fetch_kml?fset=%s&uuid=%s&kmlsize=%Lu",
                               fset->fset_name, uuidstr, kmlsize);
        CDEBUG(D_KML, "offset %llu, already %d\n", kmlsize, already);
        rc = is_remote_cgi(ctxt, name);
        g_free(name);

        g_mutex_lock(peer->srv_kml_mutex);
        if (peer->srv_kml_size > kmlsize) {
                /* We've received another go_fetch_kml request while processing
                 * this one, restart. */
                kmlsize = peer->srv_kml_size;
                goto restart;
        }
        peer->srv_go_fetch_pending = 0;
        g_mutex_unlock(peer->srv_kml_mutex);

        return rc;
}

int is_send_kmlsize(struct is_context *ctxt)
{
        char *name, *body;
        int rc, len;
        struct is_fileset *fset = ctxt->ctxt_replicator->rp_fset;
        char uuidstr[UUID_UNPARSED_SIZE];
        __u64 kmlsize;

        rc = izo_ioc_get_kml_size(fset->fset_mtpt_fd, &kmlsize);
        if (rc < 0)
                return rc;

        uuid_unparse(ctxt->ctxt_intersync->is_local_uuid, uuidstr);
        name = g_strdup_printf("set_kmlsize?fset=%s&uuid=%s&kmlsize=%Lu",
                               fset->fset_name, uuidstr, kmlsize);
        rc = is_remote_cgi(ctxt, name);
        g_free(name);
        if (rc)
                return rc;

        len = ctxt->ctxt_req->len;
        if (len == 0)
                return -1;

        body = ctxt->ctxt_req->body;
        /* remote will send '1' to indicate success */
        rc = body[0] - '1';

        return rc;
}

#if 0
int is_make_branch(struct is_context *ctxt, char *tagname, char *branchname)
{
        struct is_peer *peer = is_ctxt_peer(ctxt);
        struct is_fileset *fset = ctxt->ctxt_replicator->rp_fset;
        char *name, *body;
        int rc, len;
        char uuidstr[UUID_UNPARSED_SIZE];
        __u64 kmlsize;

        izo_ioc_get_kml_size(fset->fset_mtpt_fd, &kmlsize);

        uuid_unparse(ctxt->ctxt_intersync->is_local_uuid, uuidstr);
        name = g_strdup_printf("make_branch?fset=%s&tag=%s&branch=%s",
                               fset->fset_name, tagname, branchname);
        rc = is_remote_cgi(ctxt, name);
        g_free(name);
        if (rc < 0)
                return -1;

        len = ctxt->ctxt_req->len;
        if (len == 0)
                return -EIO;

        body = ctxt->ctxt_req->body;
        rc = body[0] - '1';

        return rc;
}
#endif

int is_branch_undo(struct is_context *ctxt, char *branch)
{
        int rc;
        struct is_peer *peer = is_ctxt_peer(ctxt);
        struct is_fileset *fset = ctxt->ctxt_replicator->rp_fset;
        char  *uri, *buf, *end;
#if 0
        int len;
        char zero_uuid[16] = {0};
        int recno, err;
        GList *kml, *kp;
        struct kml_rec *rec, *inv_rec;
#endif

        uri = g_strdup_printf("http://%s:%d/cache/.intermezzo/%s/BRANCHES/%s/kml",
                              peer->srv_name, peer->srv_port,
                              fset->fset_name, branch);

        rc = is_http_get(ctxt, uri, 0, 0);
        g_free(uri);

        if (rc || ctxt->ctxt_req == NULL || ctxt->ctxt_req->len == 0) {
                CERROR("No KML?\n");
                return -1;
        }

        buf = ctxt->ctxt_req->body;
        end = buf + ctxt->ctxt_req->len;

#if 0
        /* FIXME: I can't make any of this work.  Not a sausage.  I've also
         * seen Italian restaurants with less spaghetti. */
        kml = kml_listify(buf, end - buf);

        kp = g_list_last(kml);
        while (kp) {
                rec = g_list_nth_data(kp, 0);
                if (!recno)
                        recno = rec->suffix->recno;

                inv_rec = kml_invert(rec, ++recno);
                if (!inv_rec) {
                        char *dump = kml_print_rec(rec, 0);
                        CERROR("cannot invert record:\n%s", dump);
                        return -1;
                }

                buf = NULL;
                len = 0;
                err = kml_pack(inv_rec, &buf, &len);
                if (err) {
                        char *dump = kml_print_rec(inv_rec, 0);
                        CERROR("cannot pack inverse record: %d, %s\n%s",
                                  err, izo_error(err), dump);
                        return -1;
                }

#if 1
                /* Check packing by unpacking. */
                {
                        char *buf2 = buf;
                        char *end = buf2 + inv_rec->suffix->len;
                        char *errmsg;
                        struct kml_rec rec2;

                        err = kml_unpack(&rec2, &buf2, end);
                        if (err) {
                                char *dump = kml_print_rec(inv_rec, 0);
                                g_error("cannot unpack inverse record: %d, %s\n%s",
                                        err, izo_error(err), dump);
                                /* NOTREACHED */
                        }

                        errmsg = kml_compare(inv_rec, &rec2);
                        if (errmsg) {
                                char *dump = kml_print_rec(inv_rec, 0);
                                char *dump2 = kml_print_rec(&rec2, 0);
                                fputs(dump, stderr);
                                fputs(dump2, stderr);
                                g_error("unpack(pack(inv_rec)) != inv_rec: %s\n",
                                        errmsg);
                                /* NOTREACHED */
                        }
                }
#endif /* check packing */

                err = kml_iocreint(inv_rec, buf, buf,
                                   fset->fset_mtpt_fd, zero_uuid, 0);
                if (err) {
                        char *dump = kml_print_rec(inv_rec, 0);
                        CERROR("failure to reintegrate record: %d, %s\n%s",
                                  err, izo_error(err), dump);
                        return -1;
                }

                kml_invert_free(inv_rec);
                kp = g_list_previous(kp);
        }

        kml_invert_closes ();
#endif
        return 0;
}

int is_branch_redo(struct is_context *ctxt, char *branch)
{
        return is_get_kml_range(ctxt, branch, -1, -1);
}
