/* -*- mode: c; c-basic-offset: 8; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 *  Copyright (C) 2002 Cluster File Systems, Inc.
 *                     Author: Gordon Matzigkeit <gord@fig.org>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Compare two KML records.
 *
 */

#include <string.h>
#include <izolib.h>

/* Compare two records, returning the name of the differing field, or
   NULL for no differences. */
char *kml_compare(struct kml_rec *a, struct kml_rec *b)
{
#define cmp_var(field)                          \
        if (a->field != b->field) return #field

#define cmp_mem(field, size)                            \
do {                                                    \
        if (a->field && b->field) {                     \
                if (memcmp(a->field, b->field, size))   \
                        return #field;                  \
        } else if (a->field || b->field)                \
                return #field;                          \
} while (0)

#define cmp_memv(field, lenfield)               \
        cmp_var(lenfield);                      \
        cmp_mem(field, a->lenfield)

        cmp_mem(prefix.hdr, sizeof(struct kml_prefix_hdr));
        cmp_memv(prefix.groups, prefix.hdr->ngroups);
        cmp_var(offset);
        cmp_memv(path, pathlen);
        cmp_memv(name, namelen);
        cmp_memv(target, targetlen);
        cmp_mem(old_objectv, sizeof(struct presto_version));
        cmp_mem(new_objectv, sizeof(struct presto_version));
        cmp_mem(old_parentv, sizeof(struct presto_version));
        cmp_mem(new_parentv, sizeof(struct presto_version));
        cmp_mem(old_targetv, sizeof(struct presto_version));
        cmp_mem(new_targetv, sizeof(struct presto_version));
        cmp_var(valid);
        cmp_var(mode);
        cmp_var(uid);
        cmp_var(gid);
        cmp_var(size);
        cmp_var(mtime);
        cmp_var(ctime);
        cmp_var(flags);
        cmp_var(ino);
        cmp_var(rdev);
        cmp_var(major);
        cmp_var(minor);
        cmp_var(generation);
        cmp_mem(suffix, sizeof(struct kml_suffix));
#undef cmp_mem
#undef cmp_struct
#undef cmp_var
        return 0;
}
