/* -*- mode: c; c-basic-offset: 8; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 *  Copyright (C) 2001, 2002 Cluster File Systems, Inc.
 *                     Author: Gordon Matzigkeit <gord@fig.org>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Pack a KML buffer from a struct kml_rec.  Derived from kml_unpack.c.
 */

/* Basic recipe: Take kml_unpack, do
   s/unpack/pack/
   s/UNLOG/RELOG/
   s/char \*end/int *end/
   s/char \*\*buf/struct reptr *buf/
   s/char \*ptr = \*buf;//
   s/\*buf = ptr;//
   s/, end//
   and season to taste. */

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <linux/intermezzo_lib.h>
#include <izolib.h>

/* RELOG* are similar to LOG*, but they reallocate the pointer as
   necessary. */

struct reptr {
	char *buf;
	int len;
};

#define RELOG_CHECK_SIZE(nlen, ptr, end)                        \
	if ((ptr)->len < *end + nlen) {                         \
		(ptr)->len = ((ptr)->len * 2) ?: 32;            \
		(ptr)->buf = realloc((ptr)->buf, (ptr)->len);   \
	}

#define RELOGP(var,type,ptr,end)			\
do {							\
	RELOG_CHECK_SIZE(sizeof(type), ptr, end);	\
	memcpy((ptr)->buf + *end, var, sizeof(type));	\
	*end += sizeof(type);				\
} while (0)

#define RELOGV(var,type,ptr,end)			\
do {							\
	RELOG_CHECK_SIZE(sizeof(type), ptr, end);	\
	*((type *)((ptr)->buf + *end)) = var;		\
	*end += sizeof(type);				\
} while (0)

#define RELOGL(var,type,len,ptr,end)                                    \
do {                                                                    \
	RELOG_CHECK_SIZE(size_round(len * sizeof(type)), ptr, end);     \
	memcpy((ptr)->buf + *end, var, len * sizeof(type));             \
	*end += size_round(len * sizeof(type));                         \
} while (0)


static int kml_pack_noop(struct kml_rec *rec, struct reptr *buf, int *end)
{
	return 0;
}

 
static int kml_pack_create(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGP(rec->old_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_objectv, struct presto_version, buf, end);
	RELOGV(rec->mode, __u32, buf, end);
	RELOGV(rec->uid, __u32, buf, end);
	RELOGV(rec->gid, __u32, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	return 0;
}

 
static int kml_pack_mkdir(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGP(rec->old_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_objectv, struct presto_version, buf, end);
	RELOGV(rec->mode, __u32, buf, end);
	RELOGV(rec->uid, __u32, buf, end);
	RELOGV(rec->gid, __u32, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	return 0;
}


static int kml_pack_unlink(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGP(rec->old_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_parentv, struct presto_version, buf, end);
	RELOGP(rec->old_objectv, struct presto_version, buf, end);
        RELOGV(rec->old_mode, __u32, buf, end);
        RELOGV(rec->old_rdev, __u32, buf, end);
        RELOGV(rec->old_uid, __u64, buf, end);
        RELOGV(rec->old_gid, __u64, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGV(rec->targetlen, __u32, buf, end);
        RELOGV(rec->old_targetlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	RELOGL(rec->target, char, rec->targetlen, buf, end);
        RELOGL(rec->old_target, char, rec->old_targetlen, buf, end);
 	return 0;
}


static int kml_pack_rmdir(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGP(rec->old_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_parentv, struct presto_version, buf, end);
	RELOGP(rec->old_objectv, struct presto_version, buf, end);
        RELOGV(rec->old_mode, __u32, buf, end);
        RELOGV(rec->old_rdev, __u32, buf, end);
        RELOGV(rec->old_uid, __u64, buf, end);
        RELOGV(rec->old_gid, __u64, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGV(rec->targetlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	RELOGL(rec->target, char, rec->targetlen, buf, end);
	return 0;
}


static int kml_pack_close(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGV(rec->mode, __u32, buf, end);  // used for open_mode
	RELOGV(rec->uid, __u32, buf, end);   // used for open_uid
	RELOGV(rec->gid, __u32, buf, end);   // used for open_gid
	RELOGP(rec->old_objectv, struct presto_version, buf, end);
	RELOGP(rec->new_objectv, struct presto_version, buf, end);
	RELOGV(rec->ino, __u64, buf, end);
	RELOGV(rec->generation, __u32, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	return 0;
}


static int kml_pack_symlink(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGP(rec->new_parentv, struct presto_version, buf, end);
	RELOGP(rec->old_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_objectv, struct presto_version, buf, end);
	RELOGV(rec->uid, __u32, buf, end);
	RELOGV(rec->gid, __u32, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGV(rec->targetlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	RELOGL(rec->target, char, rec->targetlen, buf, end);
	return 0;
}


static int kml_pack_rename(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGP(rec->old_objectv, struct presto_version, buf, end);
	RELOGP(rec->new_objectv, struct presto_version, buf, end);
	RELOGP(rec->old_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_parentv, struct presto_version, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGV(rec->targetlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	RELOGL(rec->target, char, rec->targetlen, buf, end);
	return 0;
}


static int kml_pack_setattr(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGP(rec->old_objectv, struct presto_version, buf, end);
	RELOGV(rec->valid, __u32, buf, end);
	RELOGV(rec->mode, __u32, buf, end);
	RELOGV(rec->uid, __u32, buf, end);
	RELOGV(rec->gid, __u32, buf, end);
	RELOGV(rec->size, __u64, buf, end);
	RELOGV(rec->mtime, __u64, buf, end);
	RELOGV(rec->ctime, __u64, buf, end);
	RELOGV(rec->flags, __u32, buf, end);
        RELOGV(rec->old_mode, __u32, buf, end);
        RELOGV(rec->old_rdev, __u32, buf, end);
        RELOGV(rec->old_uid, __u64, buf, end);
        RELOGV(rec->old_gid, __u64, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	return 0;
}


static int kml_pack_link(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGP(rec->old_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_objectv, struct presto_version, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGV(rec->targetlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	RELOGL(rec->target, char, rec->targetlen, buf, end);
	return 0;
}


static int kml_pack_open(struct kml_rec *rec, struct reptr *buf, int *end)
{
	printf("NOT IMPLEMENTED");
	return 0;
}


static int kml_pack_mknod(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGP(rec->old_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_parentv, struct presto_version, buf, end);
	RELOGP(rec->new_objectv, struct presto_version, buf, end);
	RELOGV(rec->mode, __u32, buf, end);
	RELOGV(rec->uid, __u32, buf, end);
	RELOGV(rec->gid, __u32, buf, end);
	RELOGV(rec->major, __u32, buf, end);
	RELOGV(rec->minor, __u32, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	return 0;
}


static int kml_pack_write(struct kml_rec *rec, struct reptr *buf, int *end)
{
	printf("NOT IMPLEMENTED");
	return 0;
}


static int kml_pack_release(struct kml_rec *rec, struct reptr *buf, int *end)
{
	printf("NOT IMPLEMENTED");
	return 0;
}


static int kml_pack_trunc(struct kml_rec *rec, struct reptr *buf, int *end)
{
	printf("NOT IMPLEMENTED");
	return 0;
}


static int kml_pack_setextattr(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGP(rec->old_objectv, struct presto_version, buf, end);
	RELOGP(rec->new_objectv, struct presto_version, buf, end);
	RELOGV(rec->flags, __u32, buf, end);
	RELOGV(rec->mode, __u32, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGV(rec->namelen, __u32, buf, end);
	RELOGV(rec->targetlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	RELOGL(rec->name, char, rec->namelen, buf, end);
	RELOGL(rec->target, char, rec->targetlen, buf, end);
	return 0;
}


static int kml_pack_delextattr(struct kml_rec *rec, struct reptr *buf, int *end)
{
	RELOGP(rec->old_objectv, struct presto_version, buf, end);
	RELOGP(rec->new_objectv, struct presto_version, buf, end);
	RELOGV(rec->flags, __u32, buf, end);
	RELOGV(rec->mode, __u32, buf, end);
	RELOGV(rec->pathlen, __u32, buf, end);
	RELOGV(rec->namelen, __u32, buf, end);
	RELOGV(rec->targetlen, __u32, buf, end);
	RELOGL(rec->path, char, rec->pathlen, buf, end);
	RELOGL(rec->name, char, rec->namelen, buf, end);
	return 0;
}


static int kml_pack_kml_trunc(struct kml_rec *rec, struct reptr *buf, int *end)
{

	printf("NOT IMPLEMENTED");
	return 0;
}


typedef int (*packer)(struct kml_rec *rec, struct reptr *buf, int *end);

static packer packers[KML_OPCODE_NUM] = 
{
	[KML_OPCODE_NOOP] = kml_pack_noop,
	[KML_OPCODE_CREATE] = kml_pack_create, 
	[KML_OPCODE_MKDIR] = kml_pack_mkdir,
	[KML_OPCODE_UNLINK] = kml_pack_unlink,
	[KML_OPCODE_RMDIR] = kml_pack_rmdir,
	[KML_OPCODE_CLOSE] = kml_pack_close,
	[KML_OPCODE_SYMLINK] = kml_pack_symlink,
	[KML_OPCODE_RENAME] = kml_pack_rename,
	[KML_OPCODE_SETATTR] = kml_pack_setattr,
	[KML_OPCODE_LINK] = kml_pack_link,
	[KML_OPCODE_OPEN] = kml_pack_open,
	[KML_OPCODE_MKNOD] = kml_pack_mknod,
	[KML_OPCODE_WRITE] = kml_pack_write,
	[KML_OPCODE_RELEASE] = kml_pack_release,
	[KML_OPCODE_TRUNC] = kml_pack_trunc,
	[KML_OPCODE_SETEXTATTR] = kml_pack_setextattr,
	[KML_OPCODE_DELEXTATTR] = kml_pack_delextattr,
	[KML_OPCODE_KML_TRUNC] = kml_pack_kml_trunc
};

int kml_pack(struct kml_rec *rec, char **buf, int *len)
{
	struct reptr ptr = { *buf, *len };
	int err, end = 0;

        RELOGP(rec->prefix.hdr, struct kml_prefix_hdr, &ptr, &end);
	RELOGL(rec->prefix.groups, __u32,
               rec->prefix.hdr->ngroups, &ptr, &end);

        if (rec->prefix.hdr->opcode < 0 ||
	    rec->prefix.hdr->opcode >= sizeof(packers) / sizeof(*packers))
		return -EINVAL;
	err = packers[rec->prefix.hdr->opcode](rec, &ptr, &end);
	if (err)
		return err;
	RELOGP(rec->suffix, struct kml_suffix, &ptr, &end);

        *len = ptr.len;
        *buf = rec->buf = ptr.buf;
        rec->prefix.hdr->len = rec->suffix->len = end;

        /* Adjust the lengths in the buffer we just wrote. */
        *(__u32 *)ptr.buf = end;
        *(__u32 *)(ptr.buf + end - sizeof(__u32)) = end;

	return 0;
}
