/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster Filesystems, Inc.
 * Copyright (C) 2001 Tacit Networks, Inc. <phil@off.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <glib.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>

#include <linux/intermezzo_idl.h>

static struct protoent *tcp_proto = NULL;

int izo_resolve(char *hostname, __u64 *ip)
{
	struct hostent *h;

	/* do this asynchronously */
	h = gethostbyname(hostname);
        if (h == NULL)
                return -h_errno;

	*ip = *(unsigned int *)h->h_addr;

	return 0;
}

int is_socket_new(int nonblock)
{
        gint o;
        gint sock;
        gint rv = 0;
	gint rc;
	gint flags;

        if (!tcp_proto) {
                tcp_proto = getprotobyname("tcp"); 
                g_assert(tcp_proto);
        }

        sock = socket(PF_INET, SOCK_STREAM, tcp_proto->p_proto);
        if(sock < 0) {
                rv = errno;
                g_error("socket_new: create socket: %s", strerror(errno));
        }

        o = 1;
        if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &o, sizeof(o))) {
                rv = errno;
                g_error("Cannot set REUSEADDR socket opt");
        }
	
        if (!nonblock)
                return sock;

	flags = fcntl(sock, F_GETFL);
	if (flags == -1 ) {
		g_error("socket_new: getflags: %s", strerror(errno));
	}

	rc = fcntl(sock, F_SETFL, flags | O_NONBLOCK);
	if ( rc == -1 ) {
		g_error("cannot make socket unblocked: %s", strerror(errno));
	}

	return sock;
}
