/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster File Systems, Inc. <braam@clusterfs.com>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Upcall/Downcall handling
 *
 */

#include <glib.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <izolib.h>
#include <linux/intermezzo_lib.h>
#include <linux/intermezzo_idl.h>

static char *upcnames[] = {
        "NOOP",
        "PERMIT",
        "CONNECT",
        "GO_FETCH_KML",
        "OPEN",
        "REVOKE_PERMIT",
        "KML",
        "BACKFETCH",
        "KML_TRUNC",
        "SET_KMLSIZE",
        "BRANCH_UNDO",
        "BRANCH_REDO",
        "GET_FILEID",
        "CLIENT_MAKE_BRANCH",
        "SERVER_MAKE_BRANCH",
        "REPSTATUS",
};

char *upcall_print(struct intersync_upcall *rec, int brief)
{
        if (brief)
                return g_strdup_printf("[UPC %s : Uniq %d : PID %d : UID %d"
                                       " Path %s : Fset %s"
                                       " Offset %lld : First recno %d : Last recno %d : Async %d]\n",
                                       upcnames[rec->u_hdr->u_opc],
                                       rec->u_hdr->u_uniq,
                                       rec->u_hdr->u_pid,
                                       rec->u_hdr->u_uid,
                                       MKSTR(rec->u_path),
                                       MKSTR(rec->u_fset),
                                       rec->u_hdr->u_offset,
                                       rec->u_hdr->u_first_recno,
                                       rec->u_hdr->u_last_recno,
                                       rec->u_hdr->u_async
                                       );

        return g_strdup_printf("\n -- Upcall:\n"
                               "    Opcode     %s\n"
                               "    Uniq   %d\n" 
                               "    PID    %d\n" 
                               "    UID    %d\n"
                               "    Pathlen %d\n"
                               "    Path    %s\n"
                               "    Fsetlen %d\n"
                               "    Fset    %s\n"
                               "    Reclen  %d\n"
                               "    Offset  %lld\n"
                               "    First recno %d\n"
                               "    Last recno  %d\n"
                               "    Async      %d\n",
                               upcnames[rec->u_hdr->u_opc],
                               rec->u_hdr->u_uniq,
                               rec->u_hdr->u_pid,
                               rec->u_hdr->u_uid,
                               rec->u_hdr->u_pathlen,
                               MKSTR(rec->u_path),
                               rec->u_hdr->u_fsetlen,
                               MKSTR(rec->u_fset),
                               rec->u_hdr->u_reclen,
                               rec->u_hdr->u_offset,
                               rec->u_hdr->u_first_recno,
                               rec->u_hdr->u_last_recno,
                               rec->u_hdr->u_async
                               );
}

int upcall_unpack(struct intersync_upcall *rec, char **buf, char *end)
{
        char *ptr = *buf;

        memset(rec, 0, sizeof(*rec));
        rec->u_buf = *buf;
        UNLOGP(rec->u_hdr, struct izo_upcall_hdr, ptr, end);

        if (rec->u_hdr->u_pathlen)
                UNLOGL0(rec->u_path, char, rec->u_hdr->u_pathlen, ptr, end);
        if (rec->u_hdr->u_fsetlen)
                UNLOGL0(rec->u_fset, char, rec->u_hdr->u_fsetlen, ptr, end);
        if (rec->u_hdr->u_reclen)
                UNLOGL(rec->u_rec, char, rec->u_hdr->u_reclen, ptr, end);

        if (rec->u_hdr->u_opc < 0 || rec->u_hdr->u_opc > IZO_UPC_LARGEST_OPCODE)
                return -EINVAL;

        *buf = ptr; 
        return 0;
}

