/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster File Systems, Inc.
 *   Author: Robert Read <rread@datarithm.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *
 */

#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <glib.h>
#include <izolib.h>

/*  
 * similar to  dirname(1) 
 * returned path must be freed.
 */
char * izo_dirname(char * path) 
{
        char *dir, *s;

        s = strrchr (path, '/');
        if (s == NULL) {
                dir = g_malloc(2);
                *dir = '.';
                *(dir+1) = '\0';
        } else if (s == path) {
                dir = g_malloc(2);
                *dir = '/';
                *(dir+1) = '\0';
        } else  {
                int len = s - path;
                dir = g_malloc(len + 1);
                memcpy(dir, path, len);
                *(dir + len) = '\0';
        }
        return dir;
}

/* 
 * recursively make all the directories in a path
 * similar to mkdir -p
 */
int izo_mkpath(char * dir, int mode) 
{
        char * parent;
        struct stat stat_buf;
  
        parent = izo_dirname (dir);
        if (stat(parent, &stat_buf) < 0 && izo_mkpath(parent, mode) < 0) {
                g_free(parent);
                return -1;
        }
        g_free(parent);

        /*
         * See if it already exists.
         */
        if (stat(dir, &stat_buf) == 0) {
                if (S_ISDIR (stat_buf.st_mode)) {
                        return 0;
                }
        }
        if (mkdir (dir, mode) < 0) {
                return -1;
        }
        return 0;
}

