/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster File Systems, Inc.
 *   Author: Robert Read <rread@datarithm.net>
 *   Author: Phil Schwan <phil@clusterfs.com>
 *   Author: Peter J. Braam <braam@clusterfs.com>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <errno.h>
#include <unistd.h>

#include <izolib.h>
#include <linux/intermezzo_lib.h>

/* 
 * Return a fileid for a given user, path, fileset. The path
 * and user info is in GET_FILEID kml_rec.
 * 
 * Response is the ino and generation number packed together.
 */
int main(int argc, char **argv)
{
        char message [] =
                //"HTTP/1.1 %d\r\n"
                "Connection: Keep-Alive\r\n"
                "Content-Type: text/plain\r\n"
                "Content-Length: %d\r\n\r\n"
                "%s";
	char response[1024];
        char *docroot;
        const char *fsetname, *recstr;
        char rec[4096];
        int size;
        int fd, rc = 0;
        __u64 ino;
        __u64 generation;

        /* FIXME: does not support storing clientid mapping yet*/
 
        docroot = getenv("DOCUMENT_ROOT");
        if (docroot == NULL) {
                fprintf(stderr, "missing docroot\n");
        } else if ((fsetname = cgi_getenv("fsetname")) == NULL 
                   || strlen(fsetname) == 0) {
                fprintf(stderr, "Error: fsetname not valid.\n");
        } else if ((recstr = cgi_getenv("rec")) == NULL 
                   || strlen(recstr) == 0) {
                fprintf(stderr, "Error: rec not valid.\n");
        } else if ((size = izo_from_base64(rec, recstr)) == 0) {
                fprintf(stderr, "Error: rec decode failure.\n");
        } else if((fd = fset_cgi_open(docroot, (char *)fsetname)) <  0) {
                rc = -1;
                fprintf(stderr, "Unable to open file: %s\n", g_strerror(errno));
        } else if((rc = izo_ioc_get_fileid(fd, rec, size, &ino, &generation))) {
                fprintf(stderr, "Unable to get fileid (%d): %s\n", errno,
                        g_strerror(errno));
                close(fd);
        } else {
                close(fd);
                rc = 0;
        }
        fprintf(stderr, "rc = %d, ino %Lx, gen %Lx\n", rc, ino, generation);

        if (rc) {
                printf(message, 0, "");
        } else {
                sprintf(response, "%Lx:%Lx", ino, generation);
                printf(message, strlen(response), response);
        }

        return 0;
}
