/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster File Systems, Inc.
 *   Author: Robert Read <rread@datarithm.net>
 *   Author: Phil Schwan <phil@clusterfs.com>
 *   Author: Peter J. Braam <braam@clusterfs.com>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <errno.h>
#include <unistd.h>

#include <izolib.h>
#include <linux/intermezzo_lib.h>

#define UUID_UNPARSED_SIZE 37

static int get_fsetname(char * docroot, const char *clientid, char * fsetname, int len) 
{
        char *cache = NULL;
        struct is_config *cfg = NULL;

        cache = g_strdup_printf("%s/cache", docroot);

        cfg = read_local_config(cache);
        if (cfg == NULL) {
                g_free(cache);
                return -1;
        }
        strcpy(fsetname, cfg->cfg_fsetname);
        /* if clientid is specified, create a branch for it */
        if (clientid) {
                strcat(fsetname, "/");
                strncat(fsetname, clientid, len - strlen(fsetname));
        }

        g_free(cache);
        g_free(cfg);
        return 0;
}

static int new_uuid(char * uuidstr, int len) 
{
        uuid_t uuid;

        uuid_generate(uuid);
        uuid_unparse(uuid, uuidstr);
        return 0;
}


/* 
 * Map clientid to a specific host configuration. If the
 * clientid is not found then generate a new uuid and branch name and
 * store new mapping in clientid database.
 *
 * Query should look something like: "clientid=host-identifier" 
 * 
 * Response:
 * <profile>
 *  <uuid>UUID</uuid>
 *  <fset>fset/branch</fset>
 * </profile>
 */
int main(int argc, char **argv)
{
        char message [] =
                //"HTTP/1.1 %d\r\n"
                "Connection: Keep-Alive\r\n"
                "Content-Type: text/plain\r\n"
                "Content-Length: %d\r\n\r\n"
                "%s";
        char data_fmt [] =
                "<profile>\n"
                " <uuid>%s</uuid>\n"
                " <fset>%s</fset>\n"
                "</profile>\n";
        const char *clientid;
        char *docroot;
        int rc = -EINVAL;
        char fsetname[1024];
        char uuidstr[UUID_UNPARSED_SIZE];

        /* FIXME: does not support storing clientid mapping yet*/
        /* clientid is no longer required */
        clientid = cgi_getenv("clientid");

        docroot = getenv("DOCUMENT_ROOT");
        if (docroot == NULL) {
                fprintf(stderr, "missing docroot\n");
        } if ((rc = get_fsetname(docroot,clientid, fsetname, sizeof(fsetname)))) {
                fprintf(stderr, "Unable to read profile\n");
        } else if ((rc = new_uuid(uuidstr, sizeof(uuidstr)))) {
                fprintf(stderr, "Unable to generate uuid\n");
        } else {
                rc = 0;
        }

        if (rc) {
                printf(message, 0, "");
        } else {
                char * datastr = g_strdup_printf(data_fmt, uuidstr,
                                                 fsetname);
                printf(message, strlen(datastr), datastr);
                g_free(datastr);
        }
        return 0;
}




