/* -*- Mode: c; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 * Copyright (C) 2001, 2002 Cluster File Systems, Inc.
 * Copyright (C) 2001 Tacit Networks, Inc. <phil@off.net>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <linux/intermezzo_idl.h>
#include <linux/intermezzo_lib.h>

#include <glib.h>
#include <uuid/uuid.h>

#include <izolib.h>
/* never seen so many goto's before in my life! */
int main(int argc, char **argv)
{
	char message [] =
		//"HTTP/1.1 %d\r\n"
		"Connection: Keep-Alive\r\n"
		"Content-Type: text/plain\r\n"
		"Content-Length: %d\r\n\r\n";
        const char *fsetname, *uuidstr;
        char *p, *docroot;
        __u64 client_kmlsize;
        struct izo_rcvd_rec lr_client;
        uuid_t uuid;
        __u64 kmlsize;
        struct izo_rcvd_rec lr_server;
        char buf[sizeof(kmlsize) + sizeof(lr_server)];
        int fd, rc;


        docroot = getenv("DOCUMENT_ROOT");
        if (docroot == NULL) {
                fprintf(stderr, "missing docroot\n");
                goto error;
        }
        /* query should look something like "fset=foo&uuid=big-long-uuid-str" */
        fsetname = cgi_getenv("fset");
        uuidstr = cgi_getenv("uuid");

        if (fsetname == NULL || uuidstr == NULL) {
                CGI_ERROR("missing fset or uuid in query string\n");
                goto error;
        }

        if (uuid_parse((char *)uuidstr, uuid ) != 0) {
                fprintf(stderr, "Error: UUID \"%s\" is not valid.\n", uuidstr);
                goto error;
        }

        memcpy(&lr_client.lr_uuid, (void*)uuid, sizeof(lr_client.lr_uuid));

        if (cgi_getenv_u64("kml_size", &client_kmlsize)) {
                goto error;
        } 

        if (cgi_getenv_u64("kml_start", &lr_client.lr_local_offset)) {
                goto error;
        }

        if (cgi_getenv_u64("kml_lastseen", &lr_client.lr_remote_offset)) {
                goto error;
        }

        if (cgi_getenv_u64("kml_start_recno", &lr_client.lr_local_recno)) {
                goto error;
        }

        if (cgi_getenv_u64("kml_lastseen_recno", &lr_client.lr_remote_recno)) {
                goto error;
        }

        fd = fset_cgi_open(docroot, fsetname);
        if (fd < 0) {
                if (errno == ENOENT) 
                        CGI_ERROR("Fset %s not found, will attempt to create\n", fsetname);
                else 
                        goto error;
                fd = cache_cgi_open(docroot);
                if (fd < 0) {
                        goto error;
                }
                rc = izo_ioc_server_make_branch(fd, fsetname);
                if (rc) {
                        CGI_ERROR("unable to make branch '%s': %s\n", fsetname,
                                g_strerror(errno));
                        close(fd);
                        goto error;
                }
                close(fd);
                fd = fset_cgi_open(docroot, fsetname);
                if (fd < 0) {
                        goto error;
                }
        }
        /* call new repstatus and pass the client's offset info */
        rc = izo_ioc_repstatus(fd, client_kmlsize, &lr_client, &lr_server);
        if (rc < 0) {
                fprintf(stderr, "izo_ioc_repstatus failed, but continuing: %s\n",
                        g_strerror(errno));
                goto error;
        }

        rc = izo_ioc_get_kml_size(fd, &kmlsize);
        if (rc < 0) {
                fprintf(stderr, "izo_ioc_get_kml_size failed: %s\n",
                        g_strerror(errno));
                goto error;
        }

        p = buf;
        fprintf(stderr, "kmlsize: %Lu\n", kmlsize);
        LLOGV(kmlsize, __u64, p);
        LOGP(lr_server.lr_uuid, uuid_t, p);
        LLOGV(lr_server.lr_remote_recno, __u64, p);
        LLOGV(lr_server.lr_remote_offset, __u64, p);
        LLOGV(lr_server.lr_local_recno, __u64, p);
        LLOGV(lr_server.lr_local_offset, __u64, p);
        LLOGV(lr_server.lr_last_ctime, __u64, p);

        p = g_strdup_printf(message, sizeof(buf));
        write(1, p, strlen(p));
        g_free(p);
        write(1, buf, sizeof(buf));

        return 0;

 error:
        printf(message, 0);
        return 0;
}
