#!/usr/bin/perl
# filesystem compare tools for intermezzo.
# compare the lstat member except dev,inode,mtime,rdev
# It also compare the file content and skip the .intermezzo and lost+found
# dir
# Author:  Christopher Li  <chrisl@mountainviewdata.com>

use strict;
use File::Find;
use File::Spec;
use File::Compare;
sub filter_field {
    return ((@_)[2,3,4,5,7,9,10,11,12]);
}

@::stat_name = ("dev","inode","mode","nlink","uid","gid","rdev",
	     "size","atime","mtime","ctime","blksize","blocks");
@::chk_name = filter_field (@::stat_name);
$::ignore_empty = 0;
$::error = 0;

sub filecomp {
    my $relname = shift;
    my $fields = [];
    my $error = 0;
    map push (@{$fields},[filter_field(lstat($_))]), @_;
    for (my $i=0; $i<=$#::chk_name; $i++) {
	my @attr = map {@$_->[$i]} @$fields;
	if (grep ($attr[0]!= $_, @attr) &&
            (!$::ignore_empty || $::chk_name[$i] ne 'blocks')) { 
	    print "./",$relname,"\n" if !$error;
	    $error = 1;
	    printf "\t%s\t%s\n",$::chk_name[$i],join(" ",@attr);
	}
    }
    my $first = shift;
    if ($error==0 && -f $first && ! $::ignore_empty &&
        scalar (grep (compare($first,$_), @_))) {
	print "./",$relname,"\n\tcontent different\n";
	return 1;
    }
    return $error;
}

sub finddir {
    my $files = {};
    foreach my $dir (@_) {
	$dir = File::Spec->canonpath($dir);
	find ({
	    no_chdir => 0,
	    wanted   => sub {
		my $rel = File::Spec->abs2rel($File::Find::name,$dir);
		if  (index($rel,".intermezzo") && index($rel, "lost+found")) {
		    $files->{$rel} = $File::Find::name; 
		}
	    },
	    follow   => 0
	}, $dir);
    }
    return $files;
3}

sub dircomp {
    my $fileset = [];
    foreach my $dir (@_) {
	push @{$fileset}, finddir($dir);
    }
    foreach my $filehash (@$fileset) {
	foreach my $relname (keys %$filehash) {
	    my @fullpath = map {delete $_->{$relname}} @$fileset;
	    if (scalar(grep !defined,@fullpath)==0) {
		$::error =1 if filecomp $relname,@fullpath;
	    }else {
		$::error = 1;
		printf "%s\n\tother file missing\n",join(" | ",@fullpath);
	    }
	}
    }
    return $::error;
}

sub md5sum {
    my $file = shift;
    (my $checksum,) = split / /,`md5sum $file`;
    return $checksum;
}

sub dirlist {
    my $dir = shift;
    my $filehash = finddir($dir);
    delete $filehash->{""};
    foreach my $relname (sort(keys %$filehash)) {
	my $fullname = $filehash->{$relname};
	print "$relname ",join(" ",filter_field(lstat($filehash->{$relname})))," ",
	(-f $fullname ? md5sum($fullname):""),"\n";
    }
}

sub usage {
    print "Usage: fscomp.pl [--ignore-empty] mnt1 mnt2 [.. mntN]\n";
}

if ($ARGV[0] eq '--ignore-empty')
{
    $::ignore_empty = 1;
    shift @ARGV;
}

if (scalar (@ARGV)==0) {
    my @hostlist = ();
    for (my $i=0; $i<=$ENV{NUMCLIENT}; $i++) {
	push @hostlist, $ENV{"MNTIZO$i"};
    }
    # print join " ", @hostlist;
    exit dircomp @hostlist;
} elsif (scalar (@ARGV)==1) {
    dirlist $ARGV[0];
} elsif (scalar(@ARGV)>=2) {
    exit dircomp @ARGV;
}
