/* -*- mode: c; c-basic-offset: 8; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 *  Copyright (C) 2002 Cluster File Systems, Inc. <braam@clusterfs.com>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Kernel-based anti-reintegration of KML records
 *
 */
#include <stdio.h>
#include <glib.h>
#include <fcntl.h>
#include <stdlib.h>

#include <linux/intermezzo_fs.h>
#include <izolib.h>


int main(int argc, char **argv)
{
	int dird, err, len;
	struct kml_rec *rec, *inv_rec;
	GList *kml, *kp;
	char *buf, *errmsg, zero_uuid[16] = {0};
        int recno = 0, first_recno = 0, dry_run = 0, optind;

        optind = 1;
        if (argc - optind > 0 && !strcmp(argv[optind], "-n")) {
                optind ++;
                dry_run = 1;
        }

	if (argc - optind < 2 || argc - optind > 3) {
		printf("Usage: %s [-n] KMLFILE FSET-DIR [RECNO]\n", argv[0]);
		exit(1);
	}

        if (argc - optind > 2)
                first_recno = atoi (argv[optind + 2]);

	dird = open(argv[optind + 1], O_RDONLY);
	if (dird == -1) {
		perror("open FSET-DIR fails");
		exit(1);
	}

        kml = kml_listify_file(argv[optind], first_recno);

	kp = g_list_last(kml);
	while (kp) {
		rec = g_list_nth_data(kp, 0);
                if (!recno)
                        recno = rec->suffix->recno;

		inv_rec = kml_invert(rec, ++ recno);
		if (!inv_rec) {
			char *dump = kml_print_rec(rec, 0);
			g_error("cannot invert record:\n%s", dump);
			/* NOTREACHED */
		}

		buf = NULL;
                len = 0;
                err = kml_pack(inv_rec, &buf, &len);
                if (err) {
			char *dump = kml_print_rec(inv_rec, 0);
                        g_error("cannot pack inverse record: %d, %s\n%s",
                                err, izo_error(err), dump);
			/* NOTREACHED */
		}

#if 1
                /* Check packing by unpacking. */
                {
                        char *buf2 = buf;
                        char *end = buf2 + inv_rec->suffix->len;
                        struct kml_rec rec2;

                        err = kml_unpack(&rec2, &buf2, end);
                        if (err) {
                                char *dump = kml_print_rec(inv_rec, 0);
                                g_error("cannot unpack inverse record: %d, %s\n%s",
                                        err, izo_error(err), dump);
                                /* NOTREACHED */
                        }

                        errmsg = kml_compare(inv_rec, &rec2);
                        if (errmsg) {
                                char *dump = kml_print_rec(inv_rec, 0);
                                char *dump2 = kml_print_rec(&rec2, 0);
                                fputs(dump, stderr);
                                fputs(dump2, stderr);
                                g_error("unpack(pack(inv_rec)) != inv_rec: %s\n",
                                        errmsg);
                                /* NOTREACHED */
                        }
                }
#endif /* check packing */

                if (dry_run) {
                        /* Debug inversion. */
                        char *dump = kml_print_rec (rec, 0);
                        g_warning("orig: %s\n", dump);
                        g_free (dump);

                        dump = kml_print_rec (inv_rec, 0);
                        g_warning("invert: %s\n", dump);
                        g_free (dump);
                } else {
                        /* XXX need to send remote offset to kernel, but
                           is this it? */
                        err = kml_iocreint(inv_rec->prefix.hdr->len, 
                                           buf, rec->offset,
                                           dird, zero_uuid, 1);
                        if (err) {
                                char *dump = kml_print_rec(inv_rec, 0);
                                /* FIXME: Make this an error. */
                                g_warning("failure to reintegrate record: %d, %s\n%s",
                                          err, izo_error(err), dump);
                                /* NOTREACHED */
                        }
                }

		kml_invert_free(inv_rec);
		kp = g_list_previous(kp);
	}

        kml_invert_closes ();
	exit(0);
}
