/* -*- mode: c; c-basic-offset: 8; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 *  Copyright (C) 2001 Cluster File Systems, Inc. <braam@clusterfs.com>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Filesystem integration of KML records
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <linux/intermezzo_fs.h>
#include <glib.h>

int main(int argc, char **argv)
{
	char *buf;
	int fd, prior;
	struct stat st;
	struct kml_rec rec;
	char *ptr;
	char *end;

	if ( argc != 3 ) {
		printf("usage %s kmlfile reintdir\n", argv[0]);
		exit(1);
	}

	init_fsreintdata();

	fd = open(argv[1], O_RDONLY);
	if ( fd == -1) { 
		perror("open fails");
		exit(1);
	}

	if (fstat(fd, &st)) { 
		perror("stat fails");
		exit(1);
	}

	if (st.st_size < 1) {
		fprintf(stderr, "KML is empty\n");
		exit(1);
	}

	buf = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
	if ( !buf ) { 
		perror("mmap fails");
		exit(1);
	}

	ptr = buf;
	end = buf + st.st_size;
	prior = -1;
	while ( ptr < end ) { 
		int err;
		err = kml_unpack(&rec, &ptr, end);
		if ( err ) {
			fprintf(stderr, "unpacking: %d, %s", err, strerror(-err));
			if (prior == -1)
				g_error("failure to unpack first record\n");
			else
				g_error("failure to unpack record after %d!\n",
					prior);
		}
		prior = rec.suffix->recno;

		err = kml_fsreint(&rec, argv[2]);
		if (err) { 
                        char *dump = kml_print_rec (&rec, 0);
			g_warning("failure to reintegrate record: %d, %s\n%s",
                                  err, strerror(-err), dump);
		}
	}

	exit(0);
}
