/* -*- mode: c; c-basic-offset: 8; indent-tabs-mode: nil; -*-
 * vim:expandtab:shiftwidth=8:tabstop=8:
 *
 *  Copyright (C) 2001 Cluster File Systems, Inc. <braam@clusterfs.com>
 *
 *   This file is part of InterMezzo, http://www.inter-mezzo.org.
 *
 *   InterMezzo is free software; you can redistribute it and/or
 *   modify it under the terms of version 2 of the GNU General Public
 *   License as published by the Free Software Foundation.
 *
 *   InterMezzo is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with InterMezzo; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Printing of KML records
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>

#define UUID_UNPARSED_SIZE 37
#include <uuid/uuid.h>
#include <linux/intermezzo_fs.h>

//#include <glib.h>

int main(int argc, char **argv)
{
        char *buf, *ptr, *end, *filename;
        int fd;
        char uuidstr[UUID_UNPARSED_SIZE];
	struct stat st;
	struct izo_rcvd_rec rec;

	if ( argc != 2 ) {
		printf("usage %s last_rcvd\n", argv[0]);
		exit(1);
	}

        filename = argv[1];

        fd = open(filename, O_RDONLY);
        if (fd == -1) {
		fprintf(stderr, "Unable to open %s: %s\n", filename,
                        strerror(errno));
		exit(1);
	}

	if (fstat(fd, &st)) { 
		perror("stat fails");
		exit(1);
	}

	if (st.st_size < 1) {
		fprintf(stderr, "last_rcvd is empty\n");
		exit(1);
	}

	buf = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
	if ( !buf ) { 
		perror("mmap fails");
		exit(1);
	}
        close(fd);
	ptr = buf;
	end = buf + st.st_size;
	while ( ptr < end ) { 
                memcpy(&rec, ptr, sizeof(rec));
                uuid_unparse(rec.lr_uuid, uuidstr);
                printf("uuid                  : %s\n", uuidstr);
                printf("last remote recno     : %Lu\n", NTOH__u64(rec.lr_remote_recno));
                printf("last remote offset    : %Lu\n", NTOH__u64(rec.lr_remote_offset));
                printf("last local recno      : %Lu\n", NTOH__u64(rec.lr_local_recno));
                printf("last local offset     : %Lu\n", NTOH__u64(rec.lr_local_offset));
                printf("last backfetched ctime: %Lu\n", rec.lr_last_ctime);
                printf("\n");

                ptr += sizeof (struct izo_rcvd_rec);
        }

        if(munmap(buf, st.st_size) < 0) {
                perror("munmap fails");
                exit(1);
        }
	exit(0);
}

